# In ASE, the default energy unit is eV (electron volt).
# In xtb, the default energy unit it Eh (hartree).

# CODATA 2018 energy conversion factor

hartree2ev = 27.211386245988
hartree2kcalpermol = 627.50947337481
ev2kcalpermol = 23.060547830619026

# In mendeleev, the default length unit is pm.
# picometer (e-12) to angstrom (e-10)
pm2angstrom = 0.01